""" Script that generates a database of processed strings from the wxWindows
documentation.

These strings are in a ready format to be attached to wxPython__doc__
attributes.
"""

import htmllib, formatter, string, os
from types import *

from wxPython.wx import *
from wxPython.html import *
from wxPython.htmlhelp import *
from wxPython.grid import *
from wxPython.calendar import *
from wxPython.utils import *
from wxPython.gizmos import *
from wxPython.help import *
from wxPython.ogl import *
from wxPython.stc import *

method_warnings = []
do_warnings = 1

def ignore_method(meth):
    if meth in ('Create',):
        return true
    if len(meth) >= 3 and meth[:2] == 'On' and meth[2] in string.uppercase:
        return true
    if len(meth) and meth[0] in string.lowercase:
        return true
    if string.split(meth)[0] == 'operator':
        return true

    return false

class wxDocsIndexParser(htmllib.HTMLParser):
    """ Parse the wxWindows index file, read and transform HTML to text.

    Store only names occuring in the wxPython namespace"""
    def __init__(self, formatter, references=0, only_sigs=0, verbose=0):
        htmllib.HTMLParser.__init__(self, formatter, verbose)
        self.object = None
        self.classes = {}
        self.methods = {}
        self.functions = {}

        self.use_references = references
        self.only_sigs = only_sigs

##    def start_ul(self, attrs):
##        pass
##
##    def end_ul(self):
##        pass
##
##    def do_li(self, attrs):
##        pass

    def do_param(self, attrs):
        if self.object is not None:
            self.object[attrs[0][1]] = attrs[1][1]

    def start_object(self, attrs):
        self.object = {}

    def end_object(self):
        try:
            if self.object.has_key('Name') and self.object.has_key('Local'):
                name = self.object['Name']
                segs = string.split(name, '::')
                if len(segs) == 1:
                    try: wxClass = eval(name)
                    except Exception, err: pass
                    else:
                        if type(wxClass) is ClassType:
                            if self.use_references:
                                self.classes[name] = self.object['Local']
                            elif not self.only_sigs:
                                self.classes[name] = includeToImport(
                                                      stripHtmlAnchorBlock(
                                                       ripMainAnchorBlock(
                                                        self.object['Local'] )),
                                                      wxClass.__module__)
                elif segs[0]:
                    ripper = ripAnchorBlock
                    # Special case the constructor
                    if segs[0] == segs[1]:
                        meth = '%s.__init__'%segs[0]
                        ripper = ripConstrAnchorBlock
                    # Special case the destructor
                    elif '~'+segs[0] == segs[1]:
                        return
                        #meth = '%s.__del__'%segs[0]
                    else:
                        if ignore_method(segs[1]):
                            return
                        meth = '%s.%s'%(segs[0], segs[1])

                    try:
                        wxClass = eval(segs[0])
                    except Exception, err:
                        return

                    try:
                        wxMeth = eval(meth)
                    except:
                        # probably not the best place to check
                        # Check if it a static method -> function case
                        try:
                            eval('%s_%s'%(segs[0], segs[1]))
                        except:
                            if do_warnings:
                                method_warnings.append('warning: %s'% meth)
                        else:
                            return
                    else:
                        if type(wxMeth) is UnboundMethodType:
                            if self.use_references:
                                self.methods[meth] = self.object['Local']
                            else:
                                self.methods[meth] = pythoniseCPPSignature(
                                                       stripHtmlAnchorBlock(
                                                         ripper(
                                                           self.object['Local'] )),
                                                           self.only_sigs)


    ##            else:
    ##                try: wxFunc = eval(segs[1])
    ##                except: pass
    ##                else:
    ##                    if type(wxFunc) is BuiltinFunctionType:
    ##                        self.functions[segs[1]] = self.object['Local']
        finally:
            self.object = None

def buildDocsDatabase(indexfiles):
    res = {'classes': {},
           'methods': {},
           'functions': {}}
    for indexfile in indexfiles:
        if indexfile == '<STC>':
            p = STCHppParser()
        else:
            global wx_docs
            wx_docs = os.path.dirname(indexfile)

            w = formatter.NullWriter()
            f = formatter.NullFormatter(w)
            p = wxDocsIndexParser(f, references=0, only_sigs=0)
            p.feed(open(indexfile).read())

        res['classes'].update(p.classes)
        res['methods'].update(p.methods)
        res['functions'].update(p.functions)

    return res

def getDocsDatabase(indexfiles, outputfile):
    import marshal
    if not os.path.exists(outputfile):
        db = buildDocsDatabase(indexfiles)

        of = open(outputfile, 'wb')
        marshal.dump(db, of)
    else:
        of = open(outputfile, 'rb')
        db = marshal.load(of)
    return db

def decorateWxPython(db):
    for name, doc in db['classes'].items():
        wxClass = eval(name)
        wxClass.__doc__ = doc

    for name, doc in db['methods'].items():
        wxMeth = eval(name)
        wxMeth.im_func.__doc__ = doc

    # BuiltinFunctionType is read only
##    for name, doc in db['functions'].items():
##        wxFunc = eval(name)
##        wxFunc.__doc__ = doc

import sys
wx_docs = '???'
name_tag = '<A NAME="%s"></A>'
constr_piv = 'Default constructor.<P>'

def ripAnchorBlock(link):
    """ Rip the contents of a help page between the anchor and the next
        horizontal line """
    name, anchor = string.split(link, '#')
    data = open(wx_docs+'/'+name).read()
    block_start = string.find(data, name_tag % anchor)
    if block_start != -1:
        block_start = string.find(data, '</H3>', block_start)
        if block_start != -1:
            block_start = block_start + 5
            block_end = string.find(data, '<HR>', block_start)
            if block_end == -1:
                block_end = string.find(data, '</BODY>', block_start)
                if block_end == -1:
                    return None
            return string.replace(data[block_start:block_end], '::', '.')
    return None

def ripConstrAnchorBlock(link):
    name, anchor = string.split(link, '#')
    data = open(wx_docs+'/'+name).read()
    tag = name_tag % anchor
    block_start = string.find(data, tag)
    if block_start != -1:
        block_start = block_start + len(tag)
        block_end = string.find(data, '<HR>', block_start)
        if block_end == -1:
            block_end = string.find(data, '</BODY>', block_start)
            if block_end == -1:
                return None
        constr_start = string.find(data, constr_piv, block_start, block_end)
        if constr_start == -1:
            constr_start = string.find(data, '</H3>', block_start)
            if constr_start == -1:
                return None
            else:
                constr_start = constr_start + 5
        else:
            constr_start = constr_start + len(constr_piv)


        return string.replace(data[constr_start:block_end], '::', '.')
    else:
        print 'ripConstrAnchorBlock', link, 'not found'
    return None

def ripMainAnchorBlock(link):
    name, anchor = string.split(link, '#')
    data = open(wx_docs+'/'+name).read()
    block_start = string.find(data, name_tag % anchor)
    if block_start != -1:
        block_start = string.find(data, '<HR>', block_start)
        if block_start != -1:
            block_start = block_start + 4
            block_end = string.find(data, '<HR>', block_start)
            if block_end == -1:
                block_end = string.find(data, '</BODY>', block_start)
                if block_end == -1:
                    return None
            return string.replace(data[block_start:block_end], '::', '.')
    return None

def stripHtmlAnchorBlock(htmlblock):
    if not htmlblock:
        return ''
    import StringIO
    s = StringIO.StringIO('')
    w = formatter.DumbWriter(s)
    f = formatter.AbstractFormatter(w)
    p = htmllib.HTMLParser(f)
    p.feed(htmlblock)
    return string.strip(s.getvalue())

include = 'Include files\n\n<'
def includeToImport(helpstr, module):
    pos = string.find(helpstr, include)
    if pos != -1:
        end = string.find(helpstr, '>', pos)
        if end != -1:
            return '%sImported from\n\n<%s>%s'%(helpstr[:pos], module, helpstr[end+1:])
    return helpstr

superflueous_decorators = ['virtual', 'const']
reference_decorators = ['*', '&']

type_translation = {'void': 'None', 'wxString': 'string', 'unsigned': '',
 'char': 'int', 'long': 'int', 'NULL': 'None', 'wxList': 'list'}
def_val_translation = {'NULL': 'None', 'TRUE': 'true', 'FALSE': 'false'}
operator_translation = {'operator >=': '__gt__', 'operator &gt;=': '__gt__',
 'operator <=': '__lt__',  'operator &lt;=': '__lt__', 'operator ==': '__eq__',
 'operator +': '__add__', 'operator -': '__sub__'}

def pythoniseCPPSignature(textblock, onlyKeepSignature=0):
    """ Signature assumed to start on the first line of the textblock """
##    eol1 = string.find(textblock, '\n')
##    if eol1 == -1:
##        eol1 = len(textblock)
##    sig = textblock[:eol1]

    # The following transformations are done
    # * type translation
    # * putting the return type at the back ->
    # * swallowing void returns
    # * removal of *, &
    # * int n, choices[] -> choices: list

    # Todo
    # * Replace constructor with wxPython todo definition

    params_start = string.find(textblock, '(')
    params_end = string.find(textblock, ')', params_start)

    if params_start != -1 and params_end != -1:
        sig = textblock[:params_end+1]
        sig_props = string.split(sig[:params_start], ' ')
        for sig_p in sig_props[:]:
            if sig_p in reference_decorators:
                sig_props.remove(sig_p)
        name = sig_props[-1]
        if len(sig_props) >= 2:
            if sig_props[-2][-1] in reference_decorators:
                sig_props[-2] = sig_props[-2][:-1]
            returns = type_translation.get(sig_props[-2], sig_props[-2])
        else:
            returns = 'None'
        params = sig[params_start+1:params_end]
        param_lst = string.split(string.join(string.split(params, '\n'), ' '), ',')
        new_param_lst = []
        for param in param_lst:
            param = string.strip(param)
            if not param: continue
            items = string.split(param, ' ')
            try:
                defaultVar = items.index('=')
            except ValueError:
                defaults = ''
                nameIdx = len(items)-1
            else:
                default = items[defaultVar+1].strip()
                defaults = ' = %s' % def_val_translation.get(default, default)
                nameIdx = defaultVar - 1
            var_type = []
            for idx in range(nameIdx):
                if string.strip(items[idx]) in ('', 'const', 'unsigned'):
                    continue

                #if items[idx] == 'int*': print 'int to return val?', name, params

                if items[idx][-1] in reference_decorators:
                    item = string.strip(items[idx][:-1])
                else:
                    item = string.strip(items[idx])
                var_type.append(type_translation.get(item, item))

            param_name = items[nameIdx]
            if not param_name:
                continue

            if param_name[0] in reference_decorators:
                param_name = param_name[1:]

            if param_name == 'choices[]' and new_param_lst[-1][:2] == 'n:':
                new_param_lst[-1] = 'choices: list%s' % defaults
            else:
                new_param_lst.append('%s: %s%s'%(param_name, string.join(var_type, ' '), defaults))


        # build params no wider than code_block_width
        code_block_width = 60
        params = ['']
        for param in new_param_lst:
            if params[-1]:
                new_line = params[-1] + ', ' + param
            else:
                new_line = param
            if len(new_line) > code_block_width:
                params[-1] = params[-1] + ','
                params.append('  '+param)
            else:
                params[-1] = new_line
        params = string.join(params, '\n')

        # ignore code after last bracket
        sig_end = string.find(textblock, '\n', params_end)
        if sig_end == -1:
            sig_end = params_end + 1

        return '%s(%s)%s' % (name, params,
              returns != 'None' and ' -> '+returns or '') +\
                (not onlyKeepSignature and textblock[sig_end:] or '')
    return textblock

class GenHppParser:
    def __init__(self):
        self.object = None
        self.classes = {}
        self.methods = {}
        self.functions = {}

    def feed(self, srclines, classname):
        currentdocs = ''
        for line in srclines:
            line = line.strip()
            if line:
                if line.startswith('// '):
                    currentdocs += ('\n' + line[3:])
                else:
                    if line[-1] == ';': line = line[:-1]
                    sig = pythoniseCPPSignature(line, 1) + '\n'
                    try:
                        methname, rest = sig.split('(', 1)
                        if methname == classname:
                            methname = '__init__'
                    except ValueError:
                        print sig
                        raise
                    self.methods['%s.%s'%(classname, methname)] = \
                        sig + currentdocs
                    currentdocs = ''

class STCHppParser(GenHppParser):
    def __init__(self):
        GenHppParser.__init__(self)
        self.feed(open('wxpython/stc-docs.h').readlines(), 'wxStyledTextCtrl')

db = getDocsDatabase(('wxpython/wx/wx.hhk', 'wxpython/ogl/ogl.hhk', '<STC>',), 'wxDocStrings.msh')
decorateWxPython(db)

method_warnings.sort()
for w in method_warnings:
    print
