#----------------------------------------------------------------------
# Name:        BaseExplorer.py
# Purpose:     Controls to explore different data sources
#
# Author:      Riaan Booysen
#
# Created:     2000/11/02
# RCS-ID:      $Id: Explorer.py,v 1.7 2001/06/13 20:36:15 riaan Exp $
# Copyright:   (c) 1999, 2000 Riaan Booysen
# Licence:     GPL
#----------------------------------------------------------------------

from os import path
import os, sys
import string, time

from wxPython.wx import *

import Preferences, Utils
from Preferences import IS, wxFileDialog
from types import StringType

from wxextensions import wxxListCtrlMix
import ExplorerNodes


#(wxID_PFE, wxID_PFT, wxID_PFL) = map(lambda x: wxNewId(), range(3))

class BaseExplorerTree(wxTreeCtrl):
    def __init__(self, parent, wId, images, rootName, rootImgIdx):
        wxTreeCtrl.__init__(self, parent, wId)
        EVT_TREE_ITEM_EXPANDING(self, wId, self.OnOpen)
        EVT_TREE_ITEM_EXPANDED(self, wId, self.OnOpened)
        EVT_TREE_ITEM_COLLAPSED(self, wId, self.OnClose)
        self.SetImageList(images)

        self.rootNode = ExplorerNodes.RootNode(rootName)
        self.AddRoot(rootName, rootImgIdx, -1,
              wxTreeItemData(self.rootNode))

    def initDisplay(self):
        rootItem = self.GetRootItem()
        self.SetItemHasChildren(rootItem, true)
        self.Expand(rootItem)
        ck = 0; ti, ck = self.GetFirstChild(rootItem, ck)
        self.Expand(ti)
        self.SelectItem(ti)

    def getChildren(self):
        children = []
        cookie = 0
        selection = self.GetSelection()
        child, cookie = self.GetFirstChild(selection, cookie)
        while child.IsOk():
            children.append(child)
            child, cookie = self.GetNextChild(selection, cookie)
        return children

    def getChildrenNames(self):
        return map(lambda id, tree = self: tree.GetItemText(id), self.getChildren())

    def getChildNamed(self, node, name):
        cookie = 0
        child, cookie = self.GetFirstChild(node, cookie)
        while child.IsOk() and self.GetItemText(child) != name:
            child, cookie = self.GetNextChild(node, cookie)
        return child

    def OnOpened(self, event):
        pass

    def OnOpen(self, event):
        item = event.GetItem()
        if self.IsExpanded(item): return
        data = self.GetPyData(item)
        hasFolders = true
        if data:
            wxBeginBusyCursor()
            try:
                self.DeleteChildren(item)
                if data.viewMode == 'list':
                    lst = data.openList()
                else:
                    cols, lst = data.openList()
                hasFolders = false
                for itm in lst:
                    if itm.isFolderish():
                        hasFolders = true
                        new = self.AppendItem(item, itm.treename or itm.name,
                              itm.imgIdx, -1, wxTreeItemData(itm))
                        self.SetItemHasChildren(new, true)
                        if itm.bold: self.SetItemBold(new, true)
            finally:
                wxEndBusyCursor()

            if data.viewMode == 'grid':
                hasFolders = true

        self.SetItemHasChildren(item, hasFolders)

    def OnClose(self, event):
        item = event.GetItem()
        data = self.GetPyData(item)
        data.closeList()


class BaseExplorerList(wxListCtrl, wxxListCtrlMix):
    def __init__(self, parent, wId, filepath, pos = wxDefaultPosition, size = wxDefaultSize, updateNotify = None, style = 0):
        wxListCtrl.__init__(self, parent, wId, pos = pos, size = size, style = wxLC_LIST | wxLC_EDIT_LABELS | style)
        self.filepath = filepath
        self.idxOffset = 0
        self.updateNotify = updateNotify
        self.node = None
        self.sortData = {}

##        EVT_LIST_ITEM_SELECTED(self, wxID_PFL, self.OnItemSelect)
##        EVT_LIST_ITEM_DESELECTED(self, wxID_PFL, self.OnItemDeselect)

        self.selected = -1

        self.items = None
        self.currImages = None

    def destroy(self):
        if self.node:
            self.node.destroy()

    def selectItemNamed(self, name):
        for idx in range(self.GetItemCount()):
            item = self.GetItem(idx)
            if item.GetText() == name:
                item.SetState(wxLIST_STATE_FOCUSED | wxLIST_STATE_SELECTED)
                self.SetItem(item)
                return item

    def getSelection(self):
        # XXX Fix, this can return IndexErrors !!!
        if self.selected >= self.idxOffset:
            return self.items[self.selected-self.idxOffset]
        else:
            return None

    def getMultiSelection(self):
        """ Returns list of indexes that map back to node list """
        res = []
        for idx in range(self.idxOffset, self.GetItemCount()):
            if self.GetItemState(idx, wxLIST_STATE_SELECTED):
                res.append(idx-self.idxOffset)
        return res

    def setMode(self, mode):
        viewModes = wxLC_ICON | wxLC_LIST | wxLC_REPORT | wxLC_SMALL_ICON
        style = self.GetWindowStyleFlag() and not viewModes
        if mode == 'list':
            self.SetWindowStyle(style | wxLC_LIST)
        elif mode == 'grid':
            self.SetWindowStyle(style | wxLC_REPORT)
        else:
            raise 'Invalid mode'

    def pastelPicker(self, idx):
        return idx % 2

    def pastelise(self):
        for idx in range(1, self.GetItemCount()):
            item = self.GetItem(idx)
            if self.pastelPicker(idx):
                item.SetBackgroundColour(Preferences.pastelMedium)
            else:
                item.SetBackgroundColour(Preferences.pastelLight)
            self.SetItem(item)

    def refreshCurrent(self):
        self.refreshItems(self.currImages, self.node)

    def refreshItems(self, images, explNode):
        """ Display ExplorerNode items """
        self.selected = -1

        if self.node:
            self.node.destroy()

        self.node = explNode
        self.DeleteAllItems()

        if self.node.viewMode == 'list':
            self.setMode('list')
            self.SetImageList(images, wxIMAGE_LIST_SMALL)
            self.currImages = images

            wxBeginBusyCursor()
            try: self.items = explNode.openList()
            finally: wxEndBusyCursor()

            self.InsertImageStringItem(self.GetItemCount(), '..', explNode.upImgIdx)
            self.idxOffset = 1
            for itm in self.items:
                self.InsertImageStringItem(self.GetItemCount(),
                      itm.treename or itm.name, itm.imgIdx)

            self.filepath = explNode.resourcepath
        elif self.node.viewMode == 'grid':
            self.setMode('grid')
            self.SetImageList(None, 0)
            self.currImages = None

            wxBeginBusyCursor()
            try: self.cols, self.items = explNode.openList()
            finally: wxEndBusyCursor()

            self.addReportColumns(self.cols[1])

            idx = self.addReportItem(0, ['..'], explNode.upImgIdx)
            for item in self.items:
                idx = self.addReportItem(idx, item.row, item.imgIdx)

            if Preferences.pastels:
                self.pastelise()

        if self.updateNotify:
            self.updateNotify()

    def OnItemSelect(self, event):
        self.selected = event.m_itemIndex
        event.Skip()

    def OnItemDeselect(self, event):
        self.selected = -1
        event.Skip()

class BaseExplorerSplitter(wxSplitterWindow):
    def __init__(self, parent, modimages, rootName, rootImgIdx, editor,
          ExplorerTree, ExplorerList, titleStr, multiView = false, listHeader = false):
        wxSplitterWindow.__init__(self, parent, -1, style = wxNO_3D|wxSP_3D|wxSP_LIVE_UPDATE|wxCLIP_CHILDREN | wxNO_FULL_REPAINT_ON_RESIZE)

        self.titleStr = titleStr#'Editor - Explorer - %s'
        self.multiView = multiView
        self.listHeader = listHeader
        self.editor = editor
        self.controllers = {}

        if multiView:
            self.tabs = wxNotebook(self, -1, style = wxCLIP_CHILDREN)
            EVT_NOTEBOOK_PAGE_CHANGED(self.tabs, self.tabs.GetId(), self.OnPageChange)
            listParent = self.tabs
        else:
            listParent = self

        if listHeader:
            listParent = wxPanel(listParent, -1)
            self.bs = wxBoxSizer(wxVERTICAL)
            self.header = wxStaticText(listParent, -1, 'Header', size=(20, 20))
            self.bs.Add(self.header, 0, wxALIGN_TOP | wxGROW | wxALL, 5)

        wxID_EXPL_LIST = wxNewId()
        self.list = ExplorerList(listParent, wxID_EXPL_LIST, '',
              updateNotify = self.OnUpdateNotify, style = wxCLIP_CHILDREN | wxNO_FULL_REPAINT_ON_RESIZE)

        if listHeader:
            self.bs.Add(self.list, 1, wxEXPAND | wxALL)
            listParent.SetAutoLayout(true)
            listParent.SetSizer(self.bs)
            page = listParent
        else:
            page = self.list

        if multiView:
            self.tabs.AddPage(page, 'Items')

        self.modimages = modimages

        EVT_LEFT_DCLICK(self.list, self.OnOpen)
        EVT_KEY_UP(self.list, self.OnKeyPressed)
#        EVT_LEFT_DOWN(self.list, self.OnListClick)

        EVT_RIGHT_DOWN(self.list, self.OnListRightDown)
        EVT_RIGHT_UP(self.list, self.OnListRightUp)

        EVT_LIST_ITEM_SELECTED(self.list, wxID_EXPL_LIST, self.OnItemSelect)
        EVT_LIST_ITEM_DESELECTED(self.list, wxID_EXPL_LIST, self.OnItemDeselect)

        EVT_LIST_BEGIN_LABEL_EDIT(self, wxID_EXPL_LIST, self.OnBeginLabelEdit)
        EVT_LIST_END_LABEL_EDIT(self, wxID_EXPL_LIST, self.OnEndLabelEdit)

        wxID_EXPL_TREE = wxNewId()
        self.tree = ExplorerTree(self, wxID_EXPL_TREE, modimages, rootName, rootImgIdx)
        EVT_TREE_SEL_CHANGING(self, wxID_EXPL_TREE, self.OnSelecting)
        EVT_TREE_SEL_CHANGED(self, wxID_EXPL_TREE, self.OnSelect)

        EVT_KEY_UP(self.tree, self.OnTreeKeyPressed)

        if multiView:
            view = self.tabs
        else:
            view = self.list
        self.SplitVertically(self.tree, view, 200)
        self.SetMinimumPaneSize(20)

        self.list.SetFocus()

        self.viewPages = ['Items']
        self.viewedNodes = {}

    def addViewPage(self, page, name, node, treenode):
        self.tabs.InsertPage(1, page, name)
        wxYield()
        self.tabs.SetSelection(1)
        self.viewPages.insert(1, name)
        self.viewedNodes[name] = node, treenode

    def closeViewPage(self, idx):
        treenode = self.viewedNodes[self.viewPages[idx]][1]
        self.viewedNodes[self.viewPages[idx]][0].close()
        del self.viewedNodes[self.viewPages[idx]]
        del self.viewPages[idx]
        self.tabs.DeletePage(idx)
        self.OnUpdateNotify(treenode)

    def addTools(self, toolbar):
        if self.list.node and self.controllers.has_key(self.list.node.protocol):
            prot = self.list.node.protocol
            tbMenus = []
            for menuLst in self.controllers[prot].toolbarMenus:
                tbMenus.extend(list(menuLst))

            for wID, name, meth, bmp in tbMenus:
                if name == '-' and not bmp:
                    toolbar.AddSeparator()
                elif bmp != '-':
                    if name[0] == '+':
                        # XXX Add toggle button
                        name = name [1:]
                    Utils.AddToolButtonBmpObject(self.editor, toolbar,
                          IS.load(bmp), name, meth)

    def getMenu(self):
        if self.list.node and self.controllers.has_key(self.list.node.protocol):
            return self.controllers[self.list.node.protocol].menu
        else:
            return None

    def destroy(self):
        self.list.destroy()
        for contr in self.controllers.values():
            contr.destroy()
        del self.controllers
        del self.list
        del self.editor

    def selectTreeItem(self, item):
        node = self.tree.GetPyData(item)
        title = self.tree.GetItemText(item)
        header = node.treename
        if node:
            imgs = node.images
            if not imgs: imgs = self.modimages
            self.tabs.SetSelection(0)
            self.list.refreshItems(imgs, node)
            #wxYield()
            title = node.getTitle()
            if self.multiView:
                #print title, self.viewPages
                if title in self.viewPages:
                    header = title
                    idx = self.viewPages.index(title)
                    if idx != 1:
                        page = self.tabs.GetPage(idx)
                        self.tabs.RemovePage(idx)
                        self.tabs.InsertPage(1, page, title)
                        self.viewPages.remove(title)
                        self.viewPages.insert(1, title)
                        idx = 1
                    self.tabs.SetSelection(idx)


        self.editor.SetTitle(self.titleStr % title)
        if self.listHeader:
            self.header.SetLabel(header)


    def goUp(self):
        treeItem = self.tree.GetItemParent(self.tree.GetSelection())
        if treeItem.IsOk():
            self.tree.SelectItem(treeItem)

    def OnUpdateNotify(self, tItm = None):
        if tItm is None:
            tItm = self.tree.GetSelection()
        else:
            self.tree.SelectItem(tItm)
        if not tItm.IsOk():
            return
        # XXX this should be smarter, only refresh on folderish name change
        # XXX add or remove
        if not self.selecting and self.tree.IsExpanded(tItm):
            self.tree.Collapse(tItm)
            self.tree.Expand(tItm)
        self.editor.setupToolBar(1)

    def OnSelecting(self, event):
        self.selecting = true

    def OnSelect(self, event):
        # Event is triggered twice, work around with flag
        if self.selecting:
            item = event.GetItem()
            self.selectTreeItem(item)
        self.selecting = false

        event.Skip()

    def OnOpen(self, event):
        if self.list.selected != -1:
            name = self.list.GetItemText(self.list.selected)
            nd = self.list.node
            if name == '..':
                if not nd.openParent(self.editor):
                    self.goUp()
            else:
                item = self.list.items[self.list.selected-1]
                if item.isFolderish():
                    tItm = self.tree.GetSelection()
                    if not self.tree.IsExpanded(tItm):
                        self.tree.Expand(tItm)
                    chid = self.tree.getChildNamed(self.tree.GetSelection(), name)
                    self.tree.SelectItem(chid)
                    #if item.checkOpenInEditor():
                    #    item.open(self.editor)
                elif nd.parentOpensChildren:
                    nd.open(item, self.editor)
                else:
                    item.open(self.editor)

    def OnKeyPressed(self, event):
        key = event.KeyCode()
        if key == 13:
            self.OnOpen(event)
        elif key == 9:
            self.tree.SetFocus()
        else:
            event.Skip()

##    def OnListClick(self, event):
##        palette = self.editor.palette
##
##        if palette.componentSB.selection and self.list.node and \
##              self.list.node.canAdd(palette.componentSB.prevPage.name):
##            name, desc, Compn = palette.componentSB.selection
##            newName = self.list.node.newItem(name, Compn)
##            try:
##                self.list.refreshCurrent()
##                self.list.selectItemNamed(newName)
##            finally:
##                palette.componentSB.selectNone()
##        else:
##            event.Skip()

    def OnBeginLabelEdit(self, event):
        self.oldLabelVal = event.GetText()
        if self.list.node:
            self.list.node.notifyBeginLabelEdit(event)

    def OnEndLabelEdit(self, event):
        newText = event.GetText()
        if newText != self.oldLabelVal:
            event.Skip()
            self.list.node.renameItem(self.oldLabelVal, newText)
            self.list.refreshCurrent()

    def OnListRightDown(self, event):
        self.listPopPt = wxPoint(event.GetX(), event.GetY())

    def OnListRightUp(self, event):
        if self.list.node and self.controllers.has_key(self.list.node.protocol):
            self.list.PopupMenu(self.controllers[self.list.node.protocol].menu,
                  self.listPopPt)

    def OnItemSelect(self, event):
        self.list.OnItemSelect(event)
        if self.list.node:
            sel = self.list.getSelection()
            if not sel: sel = self.list.node
            self.editor.statusBar.setHint(sel.getDescription())

    def OnItemDeselect(self, event):
        self.list.OnItemDeselect(event)
        if self.list.node:
            self.editor.statusBar.setHint(self.list.node.getDescription())

    def OnTreeKeyPressed(self, event):
        key = event.KeyCode()
        if key == 9:
            self.list.SetFocus()
        else:
            event.Skip()

    def OnPageChange(self, event):
        self.editor.setupToolBar(viewIdx=event.GetSelection())
        event.Skip()
