#----------------------------------------------------------------------
# Name:        PaletteMapping.py
# Purpose:     Module that initialises the Palette's data and provides
#              the namespace in which design time code is evaluated
#
# Author:      Riaan Booysen
#
# Created:     1999
# RCS-ID:      $Id: PaletteMapping.py,v 1.14 2001/09/04 23:16:51 riaan Exp $
# Copyright:   (c) 1999 - 2002 Riaan Booysen
# Licence:     GPL
#----------------------------------------------------------------------

import os, glob

import Preferences, Utils
from Preferences import IS
import PaletteStore

from wxPython.wx import *

# This should be the first time the Companion classes are imported
# As the modules are imported they add themselves to the PaletteStore
from Companions.Companions import *

from Companions.GizmoCompanions import *

if Utils.IsComEnabled():
    from Companions.ComCompanions import *

from Companions.UtilCompanions import *

from Companions.DialogCompanions import *

if Utils.createAndReadConfig('Explorer').has_option('explorer', 'zope'):
    from Companions.ZopeCompanions import *

if Preferences.pluginsEnabled:
    paths = [Preferences.pyPath+ '/Plug-ins/*.py']
    if Preferences.extraPluginsPath:
        paths.append(Preferences.extraPluginsPath+'/*.py')
    if Preferences.rcPath != Preferences.pyPath and \
          os.path.isdir(Preferences.rcPath+ '/Plug-ins'):
        paths.append(Preferences.rcPath+ '/Plug-ins/*.py')

    print 'executing plugins...'
    for globpath in paths:
        for filename in glob.glob(globpath):
            try:
                execfile(filename)
            except Exception, error:
                wxLogError('Problem executing plugin %s:\n%s' %(os.path.basename(filename), str(error)) )

palette = PaletteStore.palette
newPalette = PaletteStore.newPalette
dialogPalette = PaletteStore.dialogPalette
zopePalette = PaletteStore.zopePalette
helperClasses = PaletteStore.helperClasses
compInfo = PaletteStore.compInfo

def compInfoByName(name):
    for comp in compInfo.keys():
        if comp.__name__ == name: return comp
    raise name+' not found'

def loadBitmap(name):
    """ Loads bitmap if it exists, else loads default bitmap """
    imgPath = 'Images/Palette/' + name+'.png'
    if IS.canLoad(imgPath):
        return IS.load(imgPath)
    else:
        return IS.load('Images/Palette/Component.png')


def bitmapForComponent(wxClass, wxBase='None'):
    """ Returns a bitmap for given comonent class.

    "Aquires" bitmap by traversing inheritance thru if necessary.
    """
    if wxBase != 'None': return loadBitmap(wxBase)
    else:
        cls = wxClass
        try: bse = wxClass.__bases__[0]
        except:
            if compInfo.has_key(wxClass):
                return loadBitmap(compInfo[wxClass][0])
            else:
                return loadBitmap('Component')
        try:
            while not compInfo.has_key(cls):
                cls = bse
                bse = cls.__bases__[0]

            return loadBitmap(compInfo[cls][0])
        except:
            print 'not found!'
            return loadBitmap('Component')

_NB = None
def evalCtrl(expr, localsDct=None):
    """ Function usually used to evaluate source snippets.

    Uses the namespace of this module which contain all the wxPython libs
    and also adds param localDct.
    """
    global _NB
    if not _NB:
        _NB = IS.load('Images/Inspector/wxNullBitmap.png')
    if not localsDct:
        wxNullBitmap = _NB
        localsDct = locals()
    else:
        localsDct['wxNullBitmap'] = _NB

    try:return eval(expr, globals(), localsDct)
    except:
        print expr
        raise
