#-----------------------------------------------------------------------------
# Name:        PascalSupport.py
# Purpose:     Example plugin module showing how to add new filetypes to the ide
#
# Author:      Riaan Booysen
#
# Created:     2002
# RCS-ID:      $Id: PascalSupport.py $
# Copyright:   (c) 2002
# Licence:     GPL
#-----------------------------------------------------------------------------

import os

from wxPython.wx import *
from wxPython.stc import *

import Preferences, Utils

import Controllers

import EditorHelper, EditorModels, PaletteStore
from Views import SourceViews, StyledTextCtrls

# Allocate an image index for Pascal files
EditorHelper.imgPascalModel = EditorHelper.imgCounter
EditorHelper.imgCounter = EditorHelper.imgPascalModel + 1

class PascalModel(EditorModels.SourceModel):
    modelIdentifier = 'Pascal'
    defaultName = 'pascal'  # default name given to newly created files
    bitmap = 'Pascal_s.png' # this image must exist in Images/Modules
    ext = '.pas'
    imgIdx = EditorHelper.imgPascalModel

class PascalStyledTextCtrlMix(StyledTextCtrls.LanguageSTCMix):
    def __init__(self, wId):
        # Here the settings are loaded from the file in the plugins directory
        # If the settings are copied into Boa's stc-styles.cfg, you will be able
        # to edit the Pascal styles in the STC Style Editor under Preferences
        cfgfile = Preferences.rcPath +'/Plug-ins/stc-pascal.rc.cfg'
        if not os.path.exists(cfgfile):
            cfgfile = Preferences.pyPath +'/Plug-ins/stc-pascal.rc.cfg'
        StyledTextCtrls.LanguageSTCMix.__init__(self, wId,
              (0, Preferences.STCLineNumMarginWidth), 'pascal', cfgfile)
        self.setStyles()

wxID_PASSOURCEVIEW = wxNewId()
class PascalSourceView(SourceViews.EditorStyledTextCtrl, PascalStyledTextCtrlMix):
    viewName = 'Source'
    def __init__(self, parent, model):
        SourceViews.EditorStyledTextCtrl.__init__(self, parent, wxID_PASSOURCEVIEW,
          model, (), -1)
        PascalStyledTextCtrlMix.__init__(self, wxID_PASSOURCEVIEW)
        self.active = true

wxID_PASCALCOMPILE = wxNewId()
class PascalController(Controllers.PersistentController):
    compileBmp = 'Images/Debug/Compile.png'
    def __init__(self, editor):
        Controllers.PersistentController.__init__(self, editor, PascalModel,
            DefaultViews=(PascalSourceView,),
            AdditionalViews=() )

    def addEvts(self):
        Controllers.PersistentController.addEvts(self)
        self.addEvt(wxID_PASCALCOMPILE, self.OnCompile)

    def addTools(self, toolbar, model):
        Controllers.PersistentController.addTools(self, toolbar, model)
        toolbar.AddSeparator()
        Controllers.addTool(self.editor, toolbar, self.compileBmp, 'Compile', self.OnCompile)

    def addMenus(self, menu, model):
        accls = Controllers.PersistentController.addMenus(self, menu, model)
        self.addMenu(menu, wxID_PASCALCOMPILE, 'Compile', accls, (Preferences.keyDefs['CheckSource']))
        return accls

    def OnCompile(self, event):
        wxLogWarning('Not implemented')

# Register Model for opening in the Editor
EditorHelper.modelReg[PascalModel.modelIdentifier] = PascalModel
# Register file extension
EditorHelper.extMap[PascalModel.ext] = PascalModel
# Register Controller for opening in the Editor
Controllers.modelControllerReg[PascalModel] = PascalController
# Add item to the New palette
# There needs to be a 24x24 png image:
#   Images/Palette/[Model.modelIdentifier].png
PaletteStore.paletteLists['New'].append(PascalModel.modelIdentifier)
# Link up controller for creation from the New palette
PaletteStore.newControllers[PascalModel.modelIdentifier] = PascalController
