#----------------------------------------------------------------------
# Name:        Preferences.py
# Purpose:     Global settings. Populated by resource files *.rc.py
#
# Author:      Riaan Booysen
#
# Created:     1999
# RCS-ID:      $Id: Preferences.py,v 1.24 2001/09/15 11:02:03 riaan Exp $
# Copyright:   (c) 1999 - 2002 Riaan Booysen
# Licence:     GPL
#----------------------------------------------------------------------
import os, sys, shutil, string

from wxPython import wx

#---Paths-----------------------------------------------------------------------

pyPath = sys.path[0]
if not pyPath: pyPath = os.path.abspath('')

def toPyPath(filename):
    return os.path.join(pyPath, filename)

#---Import preference namespace from resource .rc. files------------------------

print 'setting user preferences'

# To run different versions of Boa this flag forces Boa to user Preference
# settings either in the Boa root or in a .boa dir in the Boa root
if '--BlockHomePrefs' in sys.argv or '-B' in sys.argv:
    ##del os.environ['HOME']
    print 'ignoring $HOME (if set)'
    rcPath = os.path.join(pyPath, '.boa')
else:
    rcPath = os.path.join(os.environ.get('HOME', pyPath), '.boa')

# fall back to defaults in Boa src if .boa does not exist
if not os.path.isdir(rcPath):
    rcPath = pyPath

# upgrade if needed and exec in our namespace
plat = wx.wxPlatform == '__WXMSW__' and 'msw' or 'gtk'
for prefsFile, version in (('prefs.rc.py', 2),
                           ('prefs.%s.rc.py'%plat, 2),
                           ('prefskeys.rc.py', 2)):
    file = os.path.join(rcPath, prefsFile)

    # first time, install to env dir
    if not os.path.exists(file):
        shutil.copy2(os.path.join(pyPath, prefsFile), file)
    # check version
    else:
        verline = string.strip(open(file).readline())

        if len(verline) >= 14 and verline[:14] == '## rc-version:':
            rcver = int(verline[14:-2])
        else:
            rcver = 0
        if rcver < version:
            # backup and copy newest
            bkno=0;bkstr=''
            while 1:
                bkfile = os.path.splitext(file)[0]+'.py~'+bkstr
                try:
                    os.rename(file, bkfile)
                except OSError, err:
                    if err.errno != 17: raise
                    bkno=bkno+1;bkstr=str(bkno)
                else:
                    break
            shutil.copy2(os.path.join(pyPath, prefsFile), file)
            print 'Preference file %s replaced, previous version backed up to %s'%(
                  file, bkfile)

    execfile(file)

for file in ('Explorer.%s.cfg' % plat, 'stc-styles.rc.cfg'):
    rcFile = os.path.join(rcPath, file)
    if not os.path.exists(rcFile):
        shutil.copy2(os.path.join(pyPath, file), rcFile)

#---Prefs dependent on user prefs-----------------------------------------------

from ImageStore import ImageStore, ZippedImageStore

if useImageArchive and os.path.exists(pyPath+'/Images/Images.archive'):
    IS = ZippedImageStore(pyPath+'/Images', 'Images.archive', useImageCache)
else:
    IS = ImageStore(pyPath, cache=useImageCache)

import FileDlg
wxFileDialog = FileDlg.wxBoaFileDialog
del FileDlg

# If user does not override interpreter, use own interpreter path
if not pythonInterpreterPath:
    pythonInterpreterPath = sys.executable

# Ugly but temporary
if useDebugger == 'old':
    from Debugger import OldDebugger
    Debugger = OldDebugger
    del OldDebugger
elif useDebugger == 'new':
    from Debugger import Debugger

#-Window size calculations------------------------------------------------------

try:
    import win32api, win32con
except ImportError:
    # thnx Mike Fletcher
    screenWidth = int(wx.wxSystemSettings_GetSystemMetric(wx.wxSYS_SCREEN_X) - verticalTaskbarWidth)
    screenHeight = int(wx.wxSystemSettings_GetSystemMetric(wx.wxSYS_SCREEN_Y) - horizontalTaskbarHeight)
else:
    screenWidth = win32api.GetSystemMetrics(win32con.SM_CXFULLSCREEN)
    screenHeight = win32api.GetSystemMetrics(win32con.SM_CYFULLSCREEN) + 20

if wx.wxPlatform == '__WXMSW__':
    wxDefaultFramePos = wx.wxDefaultPosition
    wxDefaultFrameSize = wx.wxDefaultSize
elif wx.wxPlatform == '__WXGTK__':
    wxDefaultFramePos = (screenWidth / 4, screenHeight / 4)
    wxDefaultFrameSize = (int(round(screenWidth / 1.5)), int(round(screenHeight / 1.5)))

edWidth = screenWidth * editorScreenWidthPerc - windowManagerSide * 2
inspWidth = screenWidth - edWidth + 1 - windowManagerSide * 4
paletteHeight = paletteHeights[paletteStyle]
bottomHeight = screenHeight - paletteHeight

#-------------------------------------------------------------------------------
def cleanup():
    IS.cleanup()
    g = globals()
    cleanWxClasses = (wxColourPtr, wxSizePtr)
    for k, v in g.items():
        for Class in cleanWxClasses:
            if isinstance(v, Class):
                del g[k]
#-------------------------------------------------------------------------------
